
/*
C-code for the test for serial independence described in:
 
Diks, C. and Panchenko, V. (2007) Nonparametric Tests for Serial Independence 
Based on Quadratic Forms, Statistica Sinica, 17, 81-98 

Cees Diks & Valentyn Panchenko, May 2008
*/

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#define min(a,b) a<b ? a:b  


#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)


int N=100 /*length of series*/, Nh=5 /*number of bandwidths*/, M=3 /* emb dim */, L=1 /* lag */, ell=1;

long seed = -1;

// Uniform random number generator

double ran1(long *idum)
{
	int j;
	long k;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (*idum <= 0 || !iy) {
		if (-(*idum) < 1) *idum=1;
		else *idum = -(*idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(*idum)/IQ;
			*idum=IA*(*idum-k*IQ)-IR*k;
			if (*idum < 0) *idum += IM;
			if (j < NTAB) iv[j] = *idum;
		}
		iy=iv[0];
	}
	k=(*idum)/IQ;
	*idum=IA*(*idum-k*IQ)-IR*k;
	if (*idum < 0) *idum += IM;
	j=iy/NDIV;
	iy=iv[j];
	iv[j] = *idum;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;
}

// Standard normal random number generator

double gasdev(long *idum)
{
	double ran1(long *idum);
	static int iset=0;
	static double gset;
	double fac,rsq,v1,v2;

	if  (iset == 0) {
		do {
			v1=2.0*ran1(idum)-1.0;
			v2=2.0*ran1(idum)-1.0;
			rsq=v1*v1+v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac=sqrt(-2.0*log(rsq)/rsq);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}


// IID Standard normal data

void makeseries0(double *x, int n)
{
  int i;

  for (i=0;i!=n;i++)
    x[i] =  gasdev(&seed);

  return;
}

// NLMA, order 1

void makeseries1(double *x, int n)
{

  int i;
  double e,e1;

  e=e1 = gasdev(&seed);

  for (i=0;i!=n;i++)
  {
    e=gasdev(&seed); 
    x[i]=e+0.8*e1*e1;
    e1=e;
  } 

  return;
}

// NLMA, order 2

void makeseries2(double *x, int n)
{

  int i;
  double e,e1,e2,e3;

  e=e1=e2=e3= gasdev(&seed);

  for (i=0;i!=4;i++)
  {
    e=gasdev(&seed); 
    e3=e2;
    e2=e1;
    e1=e;
  } 

  for (i=0;i!=n;i++)
  {
    e=gasdev(&seed); 
    x[i]=e+0.6*(e1*e1+e2*e2);
    e3=e2;
    e2=e1;
    e1=e;
  } 


  return;
}

// NLMA, order 2

void makeseries3(double *x, int n)
{

  int i;
  double e,e1,e2;

  e=e1=e2= gasdev(&seed);

  for (i=0;i!=3;i++)
  {
    e=gasdev(&seed); 
    e2=e1;
    e1=e;
  } 

  for (i=0;i!=n;i++)
  {
    e=gasdev(&seed); 
    x[i]=e+0.8*e1*e2;
    e2=e1;
    e1=e;
  } 


  return;
}

//  AR(1)

void makeseries4(double *x, int n)
{
                                                                                
  int i;
                                                                                
  x[0]=gasdev(&seed);
                                                                                
  for (i=1;i!=50;i++)
  {
    x[0]=0.3*x[0]+gasdev(&seed);
  }
                                                                                
  for (i=1;i!=n;i++)
  {
    x[i]=0.3*x[i-1]+gasdev(&seed);
  }
                                                                          
                                                                            
  return;
}

// Nonlinear AR(1)

void makeseries5(double *x, int n)
{

  int i;

  x[0]=gasdev(&seed);

  for (i=1;i!=50;i++)
  {
    x[0]=0.8*pow(fabs(x[0]),0.5)+gasdev(&seed); 
  } 

  for (i=1;i!=n;i++)
  {
    x[i]=0.8*pow(fabs(x[i-1]),0.5)+gasdev(&seed); 
  } 


  return;
}


// Nonlinear AR(1)

void makeseries6(double *x, int n)
{

  int i;

  x[0]=gasdev(&seed);

  for (i=1;i!=50;i++)
  {
    x[0]=(x[0]>0 ? 1:-1)+gasdev(&seed); 
  } 

  for (i=1;i!=n;i++)
  {
    x[i]=(x[i-1]>0 ? 1:-1)+gasdev(&seed); 
  } 


  return;
}

// Bilinear

void makeseries7(double *x, int n)
{

  int i;
  double e,e1;

  x[0]=x[1]=e=e1=gasdev(&seed);

  for (i=2;i!=50;i++)
  {
    e=gasdev(&seed);    
    x[i]=0.6*e1*x[i-2]+e; 
    e1=e;
  } 

  x[0]=x[48];
  x[1]=x[49];

  for (i=2;i!=n;i++)
  {
    e=gasdev(&seed);    
    x[i]=0.6*e1*x[i-2]+e; 
    e1=e;
  } 

  return;
}


// ARCH(1)

void makeseries8(double *x, int n)
{
                                                                                                    
  int i;
  double ht;
                                                                                                    
  x[0] = gasdev(&seed);
  ht=1;
                                                                                                    
  for (i=0;i!=50;i++)
  {
    ht=1+0.4*x[0]*x[0];
    x[0]=sqrt(ht)*gasdev(&seed);
  }                                                                                             
  for (i=1;i!=n;i++)
  {
    ht=1.0+0.4*x[i-1]*x[i-1];
    x[i]=sqrt(ht)*gasdev(&seed);
  }                                                                                           
  return;
}

// GARCH(1,1)

void makeseries9(double *x, int n)
{

  int i;
  double ht;

  x[0] = gasdev(&seed);
  ht=1;

  for (i=0;i!=50;i++)
  {
    ht=0.01+0.80*ht+0.15*x[0]*x[0]; 
    x[0]=sqrt(ht)*gasdev(&seed);
  }  
  

  for (i=1;i!=n;i++)
  {
    ht=.01+0.80*ht+0.15*x[i-1]*x[i-1]; 
    x[i]=sqrt(ht)*gasdev(&seed);
  } 

  return;
}

// TAR(1)

void makeseries10(double *x, int n)
{
  int i;
  double x1;
                                                                                
  x1 = gasdev(&seed);
                                                                                
  for (i=0;i!=50;i++)
  {
    if (x1<1)
      x1 = -0.5*x1 + gasdev(&seed);
    else x1=0.4* gasdev(&seed);
  }
                                                                                
  for (i=0;i!=n;i++)
  {
    if (x1<1)
      x1 = -0.5*x1 + gasdev(&seed);
    else x1=0.4*gasdev(&seed);
    x[i] = x1;
  }
  return;
}

// Logistic map

void makeseries11(double *x, int n)
{
                                                                                                    
  int i;
  x[0]=ran1(&seed);

  for (i=1;i!=10;i++)
  {
    x[i]=4.0*x[i-1]*(1-x[i-1]);
  }

  x[0]=x[9];

  for (i=1;i!=n;i++)
  {
    x[i]=4.0*x[i-1]*(1-x[i-1]);
  }
  return;
}


// Henon

void makeseries12(double *x, int n)
{

  int i;

  x[0] = gasdev(&seed)/10;
  x[1] = gasdev(&seed)/3;

  for (i=2;i!=30;i++)
    x[i]=1+0.3*x[i-2]-1.4*x[i-1]*x[i-1];
  
  x[0]=x[i-2];
  x[1]=x[i-1];    
    
  for (i=2;i!=n;i++)
    x[i]=1+0.3*x[i-2]-1.4*x[i-1]*x[i-1];
  return;
}



// Noisy Henon

void makeseries13(double *x, int n)
{

  int i;

  x[0] = gasdev(&seed)/10;
  x[1] = gasdev(&seed)/3;

  for (i=2;i!=30;i++)
    x[i]=1+0.3*x[i-2]-1.4*x[i-1]*x[i-1];

  x[0]=x[i-2];
  x[1]=x[i-1];

  for (i=2;i!=n;i++)
    x[i]=1+0.3*x[i-2]-1.4*x[i-1]*x[i-1];
  
  normalise(x,n);

  for (i=0;i!=n;i++)
    x[i]=x[i]+.2*gasdev(&seed);

  return;
}




normalise(double *x, int n)
{

  int i;
  double av, av_sq, tmp;

/* Determine mean and variance */

    av = 0.0;
    av_sq = 0.0;
    for(i=0;i!=n;i++) {
      av += x[i];
      av_sq += x[i]*x[i];
    }
    av /= (double)(n);
    av_sq /= (double)(n);

/* renormalise the time series */

    tmp = sqrt((av_sq - av*av)*(double)(N)/(double)(N-1));

    for(i=0;i!=N;i++)
      x[i] /= tmp;

  return;
}

absol(double *x, int n)
{
  int i;
    for(i=0;i!=n;i++) 
      x[i]=fabs(x[i]);
}                                                                                                    


int sqdist_cauchy(double Q[], double x[N], double *h)
{
  int i, j, k, s, Nvec;
  double v[Nh], tmp, C1[Nh][N], w[Nh], u[Nh];
  
  Nvec = N-(M-1)*L;
  
  for (k=0;k!=Nh;k++)
  {
    v[k] = h[k]*h[k];
    Q[k] = 0.0;
  }
  
  
  for (i=0;i!=N;i++)
  {
    for (k=0;k!=Nh;k++)
      C1[k][i] = 0;
    
    for (j=0;j!=N;j++)
    { 
      tmp = (x[i]-x[j])*(x[i]-x[j]);
      for (k=0;k!=Nh;k++)
        C1[k][i] += (1.0/(1.0+tmp/v[k]))/(N);
    }
  }

  
  for (i=0;i!=Nvec;i++)
  {
    for (j=0;j!=Nvec;j++)
    {
      tmp = 0.0;
      for (k=0;k!=Nh;k++)
	u[k] = 1.0;

      for (s=0;s<M*L;s=s+L)
      {
        tmp = (x[i+s]-x[j+s])*(x[i+s]-x[j+s]);
        for (k=0;k!=Nh;k++)
	  u[k] *= (1.0/(1.0+tmp/v[k]));
      }

      for (k=0;k!=Nh;k++)
        Q[k] += u[k]/(Nvec*Nvec);
    }

    for (k=0;k!=Nh;k++)
    {
      w[k] = 1.0;
      for (s=0;s<M*L;s=s+L)
	w[k] *= C1[k][i+s];
      Q[k] -= 2.0*w[k]/Nvec;
    }
  }


  return(0);
}

int sqdist_doubexp(double Q[], double x[N], double *h)
{
  int i, j, k, s, Nvec;
  double v[Nh], tmp, C1[Nh][N], w[Nh];
  
  Nvec = N-(M-1)*L;
  
  for (k=0;k!=Nh;k++)
  {
    v[k] = 4.0*h[k];
    Q[k] = 0.0;
  }
  
  
  for (i=0;i!=N;i++)
  {
    for (k=0;k!=Nh;k++)
      C1[k][i] = 0.0;
    
    for (j=0;j!=N;j++)
    { 
      tmp = fabs(x[i]-x[j]);
      for (k=0;k!=Nh;k++)
        C1[k][i] += exp(-tmp/v[k])/N;
    }
  }

  
  for (i=0;i!=Nvec;i++)
  {
    for (j=0;j!=Nvec;j++)
    {
	tmp = 0.0;
	
	for (s=0;s<M*L;s=s+L)
          tmp += fabs(x[i+s]-x[j+s]);
	
        for (k=0;k!=Nh;k++)
	  Q[k] += exp(-tmp/v[k])/(Nvec*Nvec);
    }

    for (k=0;k!=Nh;k++)
    {
      w[k] = 1.0;
      for (s=0;s<M*L;s=s+L)
	w[k] *= C1[k][i+s];
      
      Q[k] -= 2.0*w[k]/Nvec;
    }

  }

  return(0);
}

int sqdist_gauss(double Q[], double x[N], double *h)
{
  int i, j, k, s, Nvec;
  double v[Nh], tmp, C1[Nh][N], w[Nh];
  
  Nvec = N-(M-1)*L;
  
  for (k=0;k!=Nh;k++)
  {
    v[k] = 4.0*h[k]*h[k];
    Q[k] = 0.0;
  }
  
  
  for (i=0;i!=N;i++)
  {
    for (k=0;k!=Nh;k++)
      C1[k][i] = 0.0;
    
    for (j=0;j!=N;j++)
    { 
      tmp = (x[i]-x[j])*(x[i]-x[j]);
      for (k=0;k!=Nh;k++)
        C1[k][i] += exp(-tmp/v[k])/N;
    }
  }

  
  for (i=0;i!=Nvec;i++)
  {
    for (j=0;j!=Nvec;j++)
    {
	tmp = 0.0;
	
	for (s=0;s<M*L;s=s+L)
          tmp += (x[i+s]-x[j+s])*(x[i+s]-x[j+s]);
	
        for (k=0;k!=Nh;k++)
	  Q[k] += exp(-tmp/v[k])/(Nvec*Nvec);
    }

    for (k=0;k!=Nh;k++)
    {
      w[k] = 1.0;
      for (s=0;s<M*L;s=s+L)
	w[k] *= C1[k][i+s];
      
      Q[k] -= 2.0*w[k]/Nvec;
    }

  }

  return(0);
}



void shuffle(double x1[N], double x[N])
{
  int i, k;
  double tmp;

  for (i=0;i!=N;i++)
    x1[i] = x[i];
  
  for (i=0;i!=N-1;i++)
  {
    k = i + (int)(ran1(&seed)*(N-i));
    tmp = x1[i];
    x1[i] = x1[k];
    x1[k] = tmp;
  }

  return;
}


void InsertionSort(double *X, int *S, int n)
{
    int i, *I;
    int j;
    int r;
    double R;

    I= (int*) malloc (n*sizeof(int));

    for (i=0;i<n;i++)
      I[i]=i;

    for (i=1; i<n; i++)
      {
        R = X[i];
        r = i;
	for (j=i-1; (j>=0) && (X[j]>R); j--)
        {
	  X[j+1] = X[j];
          I[j+1] = I[j];
        }
	X[j+1] = R;
        I[j+1] = r;
      }
    for (i=0; i<n; i++)
      S[I[i]]=i;

}


void  uniform (double *X, int n)
{
  int *I, i;

  I = (int*) malloc (n*sizeof(int));
  InsertionSort(X, I, n);

  for (i=0;i<n;i++){
    X[i] = (double) (I[i]+1)/(n+1);
    //    fprintf(stdout,"%f ",X[i]);
  }
}

void  cauchy (double *X, int n)
{
  int i;
  double b=1.0; //full width at the half maximum
  

  uniform(X,n);

  for (i=0;i<n;i++){
    X[i] = b*tan(M_PI*(X[i]-0.5)); //+median
    //    fprintf(stdout,"%f ",X[i]);
  }
}



int main(int argc, char **argv)
{

  int Nsim=1000 /*# simulations */, B=100 /* bootstrap replications + 1 */, Mod; /* model */
  /* double hmin=0.5, hmax=2.0; */
  double hmin = 0.5, hmax = 2.0;
  char Ker='g', Trans='n';  

  int id, err;

  if (argc>1)
  {
    M=atoi(argv[1]);
    if (argc>2)
    {
      L=atoi(argv[2]);
      if (argc>3)
      {
        Ker=argv[3][0];
        if (argc>4)
        {
          Trans = atoi(argv[4]);
          if (argc>5)
	  {
            B=atoi(argv[5]);
            if (argc>6)
            {
              Nsim=atoi(argv[6]);
              if (argc>7)
              {
                Nh=atoi(argv[7]);
                if (argc>8)
	        {
                  hmin=atof(argv[8]);
                  hmax=atof(argv[9]);
                }
              }
            }
          }
        }
      }
    } 
  }

  int i, j, k, nsim, b, b1, I[Nsim], ptail, Ttail, nN, tr;
  double h[Nh], power, powerh[Nh],pval[B][Nh], Tpval, Q[B][Nh], T[B], msg[Nh+1],msgsum[Nh+1];
  double x[N], x1[N];

  hmin/=hmax;
  for (k=0;k!=Nh;k++)
  {
    h[k] = hmax*pow(hmin,k/(double)(Nh-1));
    //fprintf(stdout,"h=%f\n",h[k]);    
  }
  hmin*=hmax;

  for (Mod=0;Mod<14;Mod++)	  
  {
    //  for (tr=0;tr<1;tr++)
    // {
    //    if (tr==0) Ker='g';
    //    else if (tr==1) Ker='d';
    //    else if (tr==2) Ker='c';
					 
    fprintf(stderr,"\nModel: %d\n",Mod);
 
	  
  for (k=0;k!=Nh;k++)
    powerh[k]=0.0;
  power=0.0;
    
  seed=-id-1;		  
  for (nsim=0;nsim<Nsim;nsim++)
  {
    fprintf(stderr,"\r nsim = %d",nsim);

    switch (Mod)
    {
    
    case  0:
      makeseries0(x,N);
      break; 

    case  1:
      makeseries1(x,N);
      break; 

    case  2:
      makeseries2(x,N);
      break; 

    case  3:
      makeseries3(x,N);
      break; 

    case  4:
      makeseries4(x,N);
      break; 

    case  5:
      makeseries5(x,N);
      break; 

    case  6:
      N=50;
      makeseries6(x,N);
      break;
 
    case  7:
      N=100;
      makeseries7(x,N);
      break;
 
    case  8:
      makeseries8(x,N);
      break; 

    case  9:
      makeseries9(x,N);
      break;

    case  10:
      makeseries10(x,N);
      break; 

    case  11:
      N=100;
      makeseries11(x,N);
      N=20;
      break;

    case  12:
      N=100;
      makeseries12(x,N);
      N=20;
      break;

    case  13:
      N=100;
      makeseries13(x,N);
      N=20;
      break;
    }

    switch (Trans)
    {
    case  'u':
      uniform(x,N);
      break;
    case  'n':
      normalise(x,N);
      break;
    case  'a':
      absol(x,N); normalise(x,N);
      break;
    case  'c':
      cauchy(x,N); normalise(x,N);
      break;
    }

    switch (Ker)
    {
    case  'c':
      sqdist_cauchy(Q[0],x,h);
      break; 
    case  'd':
      sqdist_doubexp(Q[0],x,h);
      break;
    case  'g':
      sqdist_gauss(Q[0],x,h);
      break;
    }


    for (b=1;b!=B;b++)
    {
      shuffle(x1,x);

      switch (Ker)
      {
      case  'c':
        sqdist_cauchy(Q[b],x1,h);
        break; 
      case  'd':
        sqdist_doubexp(Q[b],x1,h);
        break;
      case  'g':
        sqdist_gauss(Q[b],x1,h);
        break;
      }


    }

    Ttail=Tpval=0;

    /* determine p & Tp values */
    for (b=0;b<B;b++)
    {
      T[b]=1;
      for (k=0;k<Nh;k++)

      {
        pval[b][k] = 0;
        ptail=0;

        for (b1=0;b1<B;b1++)
        {

	  if (Q[b][k] < Q[b1][k]) pval[b][k]++;
	  else if (Q[b][k] == Q[b1][k]) ptail++;

        }
       /* randomization @ tails*/
        pval[b][k] += ceil(ptail*ran1(&seed));
        pval[b][k] /= (double)(B);
        
        if (b==0 && (pval[0][k]<=0.05)) 
          powerh[k]++;

        //fprintf(stdout,"%i,%i p %f Q %e ptail %i power %f\n",b,k,pval[b][k],Q[b][k],ptail,powerh[k]);
        T[b]=min(T[b],pval[b][k]);
      }

      //fprintf(stdout,"%f\n",T[b]);
      if (T[0] > T[b])	Tpval++;
      else if (T[0] == T[b]) Ttail++; 
    }   

    Tpval += ceil(Ttail*ran1(&seed));
    Tpval /= (double)(B);

    if (Tpval<=0.05) 
      power++;

    //    fprintf(stderr,"%f\n",Tpval);

 } //of sim

  power/=(double)(Nsim);

  fprintf(stdout, "\nOverall power: %f\n",power);

  fprintf(stdout,"\nModel Kernel Transform N Nh minh maxh Nsim B M L Power h[i] Power[i]\n");
  
    for(k=0;k<Nh;k++)
    {
      powerh[k] /= (double)(Nsim);
      fprintf(stdout, "%d %c %c %d %d %f %f %d %d %d %d %f %f %f\n",Mod,Ker,Trans,N,Nh,hmin,hmax,Nsim,B,M,L,power,h[k],powerh[k]); 
    }
  
  }

  return(0);
}
