/*
C-code for the test for serial independence described in:
 
Diks, C. and Panchenko, V. (2007) Nonparametric Tests for Serial Independence 
Based on Quadratic Forms, Statistica Sinica, 17, 81-98 

Cees Diks & Valentyn Panchenko, May 2008
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#define min(a,b) a<b ? a:b  


#define IA 16807
#define IM 2147483647
#define AM (1.0/IM)
#define IQ 127773
#define IR 2836
#define NTAB 32
#define NDIV (1+(IM-1)/NTAB)
#define EPS 1.2e-7
#define RNMX (1.0-EPS)


int N, Nh=5 /*number of bandwidths*/, M=3 /* emb dim */, L=1; /* lag */

long seed = -1;

// Uniform random number generator

double ran1(long *idum)
{
	int j;
	long k;
	static long iy=0;
	static long iv[NTAB];
	double temp;

	if (*idum <= 0 || !iy) {
		if (-(*idum) < 1) *idum=1;
		else *idum = -(*idum);
		for (j=NTAB+7;j>=0;j--) {
			k=(*idum)/IQ;
			*idum=IA*(*idum-k*IQ)-IR*k;
			if (*idum < 0) *idum += IM;
			if (j < NTAB) iv[j] = *idum;
		}
		iy=iv[0];
	}
	k=(*idum)/IQ;
	*idum=IA*(*idum-k*IQ)-IR*k;
	if (*idum < 0) *idum += IM;
	j=iy/NDIV;
	iy=iv[j];
	iv[j] = *idum;
	if ((temp=AM*iy) > RNMX) return RNMX;
	else return temp;
}

// Standard normal random number generator

double gasdev(long *idum)
{
	double ran1(long *idum);
	static int iset=0;
	static double gset;
	double fac,rsq,v1,v2;

	if  (iset == 0) {
		do {
			v1=2.0*ran1(idum)-1.0;
			v2=2.0*ran1(idum)-1.0;
			rsq=v1*v1+v2*v2;
		} while (rsq >= 1.0 || rsq == 0.0);
		fac=sqrt(-2.0*log(rsq)/rsq);
		gset=v1*fac;
		iset=1;
		return v2*fac;
	} else {
		iset=0;
		return gset;
	}
}



normalise(double *x, int n)
{

  int i;
  double av, av_sq, tmp;

/* Determine mean and variance */

    av = 0.0;
    av_sq = 0.0;
    for(i=0;i!=n;i++) {
      av += x[i];
      av_sq += x[i]*x[i];
    }
    av /= (double)(n);
    av_sq /= (double)(n);

/* renormalise the time series */

    tmp = sqrt((av_sq - av*av)*(double)(n)/(double)(n-1));

    for(i=0;i!=n;i++)
      x[i] /= tmp;

  return;
}

absol(double *x, int n)
{
  int i;
    for(i=0;i!=n;i++) 
      x[i]=fabs(x[i]);
}                                                                                                    


int sqdist_cauchy(double Q[], double *x, int N, double *h)
{
  int i, j, k, s, Nvec;
  double v[Nh], tmp, C1[Nh][N], w[Nh], u[Nh];
  
  Nvec = N-(M-1)*L;
  
  for (k=0;k!=Nh;k++)
  {
    v[k] = h[k]*h[k];
    Q[k] = 0.0;
  }
  
  
  for (i=0;i!=N;i++)
  {
    for (k=0;k!=Nh;k++)
      C1[k][i] = 0;
    
    for (j=0;j!=N;j++)
    { 
      tmp = (x[i]-x[j])*(x[i]-x[j]);
      for (k=0;k!=Nh;k++)
        C1[k][i] += (1.0/(1.0+tmp/v[k]))/(N);
    }
  }

  
  for (i=0;i!=Nvec;i++)
  {
    for (j=0;j!=Nvec;j++)
    {
      tmp = 0.0;
      for (k=0;k!=Nh;k++)
	u[k] = 1.0;

      for (s=0;s<M*L;s=s+L)
      {
        tmp = (x[i+s]-x[j+s])*(x[i+s]-x[j+s]);
        for (k=0;k!=Nh;k++)
	  u[k] *= (1.0/(1.0+tmp/v[k]));
      }

      for (k=0;k!=Nh;k++)
        Q[k] += u[k]/(Nvec*Nvec);
    }

    for (k=0;k!=Nh;k++)
    {
      w[k] = 1.0;
      for (s=0;s<M*L;s=s+L)
	w[k] *= C1[k][i+s];
      Q[k] -= 2.0*w[k]/Nvec;
    }
  }


  return(0);
}

int sqdist_doubexp(double Q[], double *x, int N, double *h)
{
  int i, j, k, s, Nvec;
  double v[Nh], tmp, C1[Nh][N], w[Nh];
  
  Nvec = N-(M-1)*L;
  
  for (k=0;k!=Nh;k++)
  {
    v[k] = 4.0*h[k];
    Q[k] = 0.0;
  }
  
  
  for (i=0;i!=N;i++)
  {
    for (k=0;k!=Nh;k++)
      C1[k][i] = 0.0;
    
    for (j=0;j!=N;j++)
    { 
      tmp = fabs(x[i]-x[j]);
      for (k=0;k!=Nh;k++)
        C1[k][i] += exp(-tmp/v[k])/N;
    }
  }

  
  for (i=0;i!=Nvec;i++)
  {
    for (j=0;j!=Nvec;j++)
    {
	tmp = 0.0;
	
	for (s=0;s<M*L;s=s+L)
          tmp += fabs(x[i+s]-x[j+s]);
	
        for (k=0;k!=Nh;k++)
	  Q[k] += exp(-tmp/v[k])/(Nvec*Nvec);
    }

    for (k=0;k!=Nh;k++)
    {
      w[k] = 1.0;
      for (s=0;s<M*L;s=s+L)
	w[k] *= C1[k][i+s];
      
      Q[k] -= 2.0*w[k]/Nvec;
    }

  }

  return(0);
}

int sqdist_gauss(double Q[], double *x, int N, double *h)
{
  int i, j, k, s, Nvec;
  double v[Nh], tmp, C1[Nh][N], w[Nh];
  
  Nvec = N-(M-1)*L;
  
  for (k=0;k!=Nh;k++)
  {
    v[k] = 4.0*h[k]*h[k];
    Q[k] = 0.0;
  }
  
  
  for (i=0;i!=N;i++)
  {
    for (k=0;k!=Nh;k++)
      C1[k][i] = 0.0;
    
    for (j=0;j!=N;j++)
    { 
      tmp = (x[i]-x[j])*(x[i]-x[j]);
      for (k=0;k!=Nh;k++)
        C1[k][i] += exp(-tmp/v[k])/N;
    }
  }

  
  for (i=0;i!=Nvec;i++)
  {
    for (j=0;j!=Nvec;j++)
    {
	tmp = 0.0;
	
	for (s=0;s<M*L;s=s+L)
          tmp += (x[i+s]-x[j+s])*(x[i+s]-x[j+s]);
	
        for (k=0;k!=Nh;k++)
	  Q[k] += exp(-tmp/v[k])/(Nvec*Nvec);
    }

    for (k=0;k!=Nh;k++)
    {
      w[k] = 1.0;
      for (s=0;s<M*L;s=s+L)
	w[k] *= C1[k][i+s];
      
      Q[k] -= 2.0*w[k]/Nvec;
    }

  }

  return(0);
}



void shuffle(double *x1, double *x, int N)
{
  int i, k;
  double tmp;

  for (i=0;i!=N;i++)
    x1[i] = x[i];
  
  for (i=0;i!=N-1;i++)
  {
    k = i + (int)(ran1(&seed)*(N-i));
    tmp = x1[i];
    x1[i] = x1[k];
    x1[k] = tmp;
  }

  return;
}


void InsertionSort(double *X, int *S, int n)
{
    int i, *I;
    int j;
    int r;
    double R;

    I= (int*) malloc (n*sizeof(int));

    for (i=0;i<n;i++)
      I[i]=i;

    for (i=1; i<n; i++)
      {
        R = X[i];
        r = i;
	for (j=i-1; (j>=0) && (X[j]>R); j--)
        {
	  X[j+1] = X[j];
          I[j+1] = I[j];
        }
	X[j+1] = R;
        I[j+1] = r;
      }
    for (i=0; i<n; i++)
      S[I[i]]=i;

}


void  uniform (double *X, int n)
{
  int *I, i;

  I = (int*) malloc (n*sizeof(int));
  InsertionSort(X, I, n);

  for (i=0;i<n;i++){
    X[i] = (double) (I[i]+1)/(n+1);
    //    fprintf(stdout,"%f ",X[i]);
  }
}

void  cauchy (double *X, int n)
{
  int i;
  double b=1.0; //full width at the half maximum
  

  uniform(X,n);

  for (i=0;i<n;i++){
    X[i] = b*tan(M_PI*(X[i]-0.5)); //+median
    //    fprintf(stdout,"%f ",X[i]);
  }
}



int main(int argc, char **argv)
{
  int B=100; /* bootstrap replications + 1 */
  /* set min and max bandwidth */
  double hmin = 0.5, hmax = 2.0;
  char Ker='g', Trans='n';
  char filename[128]="";
  FILE *infil;

  int id, err;

  if (argc>1)
  {
    strcpy(filename,argv[1]);
    fprintf(stderr,"Opening file %s\n",filename);
    if (argc>2)
    {
      M=atoi(argv[2]);
      if (argc>3)
      {
        L=atoi(argv[3]);
        if (argc>4)
        {
          Ker=argv[4][0];
          if (argc>5)
          {
            Trans = atoi(argv[5]);
            if (argc>6)
	    {
              B=atoi(argv[6]);
              if (argc>7)
              {
                Nh=atoi(argv[7]);
                if (argc>8)
  	        {
                  hmin=atof(argv[8]);
                  hmax=atof(argv[9]);
                }
              }
            }
          }
        }
      }
    } 
  }

  int i, j, k, nsim, b, b1, I, ptail, Ttail, nN, tr;
  double h[Nh], pval[B][Nh], Tpval, Q[B][Nh], T[B], msg[Nh+1], msgsum[Nh+1];
  double *x, *x1, tmp;

  

/* ----- Read data from file ------- */


  if ((infil = fopen(filename,"r"))==NULL)
  {
     fprintf(stderr,"File reading error. Exiting...\n");
     exit(1);
  }

  i=0;

  while (fscanf(infil,"%lf",&tmp)!=EOF)
  {
    i++;
  }

  fclose(infil);

  fprintf(stderr,"%d values read\n",i);
  
  x = (double *) malloc(i*sizeof(double));
  x1 = (double *) malloc(i*sizeof(double));


  i=0;
  if ((infil = fopen(filename,"r"))==NULL)
  {
     fprintf(stderr,"File reading error. Exiting...\n");
  }
  else 
  {
    while (fscanf(infil,"%lf",&x[i])!=EOF)
      i++;
  }

  N = i;

/* ------------------------------------ */

  hmin/=hmax;

  fprintf(stdout,"Bandwidth values\n");    
  for (k=0;k!=Nh;k++)
  {
    h[k] = hmax*pow(hmin,k/(double)(Nh-1));
    fprintf(stdout," h = %f\n",h[k]);    
  }
  hmin*=hmax;

    switch (Trans)
      {
    case  'u':
      uniform(x,N);
      break;
    case  'n':
      normalise(x,N);
      break;
    case  'a':
      absol(x,N); normalise(x,N);
      break;
    case  'c':
      cauchy(x,N); normalise(x,N);
      break;
    }

    switch (Ker)
    {
    case  'c':
      sqdist_cauchy(Q[0],x,N,h);
      break; 
    case  'd':
      sqdist_doubexp(Q[0],x,N,h);
      break;
    case  'g':
      sqdist_gauss(Q[0],x,N,h);
      break;
    }


    for (b=1;b!=B;b++)
    {
      shuffle(x1,x,N);

      switch (Ker)
      {
      case  'c':
        sqdist_cauchy(Q[b],x1,N,h);
        break; 
      case  'd':
        sqdist_doubexp(Q[b],x1,N,h);
        break;
      case  'g':
        sqdist_gauss(Q[b],x1,N,h);
        break;
      }


    }

    Ttail=Tpval=0;

    /* determine p & Tp values */
    for (b=0;b<B;b++)
    {
      T[b]=1;
      for (k=0;k<Nh;k++)

      {
        pval[b][k] = 0;
        ptail=0;

        for (b1=0;b1<B;b1++)
        {

	  if (Q[b][k] < Q[b1][k]) pval[b][k]++;
	  else if (Q[b][k] == Q[b1][k]) ptail++;

        }
       /* randomization @ tails*/
        pval[b][k] += ceil(ptail*ran1(&seed));
        pval[b][k] /= (double)(B);
        

        //fprintf(stdout,"%i,%i p %f Q %e ptail %i power %f\n",b,k,pval[b][k],Q[b][k],ptail,powerh[k]);
        T[b]=min(T[b],pval[b][k]);
      }

      //      fprintf(stdout,"%f\n",T[b]);
      if (T[0] > T[b])	
        Tpval++;
      else if (T[0] == T[b]) 
        Ttail++; 
    }   

    Tpval += ceil(Ttail*ran1(&seed));
    Tpval /= (double)(B);
    fprintf(stderr,"p-value: %f\n",Tpval);


  return(0);
}
